@@ -43,6 +43,9 @@
 #include <gtk/gtkdnd.h>
 #include <gtk/gtkentry.h>
 #include <gtk/gtkfilesel.h>
+#include <gtk/gtkfilechooserdialog.h>
+#include <gtk/gtkstock.h>
+#include <gtk/gtkdialog.h>
 #include <gtk/gtkmain.h>
 #include <gtk/gtksignal.h>
 
@@ -65,6 +68,8 @@ struct _GnomeFileEntryPrivate {
 
 	gboolean directory_entry : 1; /*optional flag to only do directories*/
 
+	gboolean use_filechooser : 1;
+
 	/* FIXME: Non local files!! */
 	/* FIXME: executable_entry as used in gnome_run */
 	/* FIXME: multiple_entry for entering multiple filenames */
@@ -106,7 +111,8 @@ enum {
 	PROP_FILENAME,
 	PROP_DEFAULT_PATH,
 	PROP_GNOME_ENTRY,
-	PROP_GTK_ENTRY
+	PROP_GTK_ENTRY,
+	PROP_USE_FILECHOOSER
 };
 
 /* Note, can't use boilerplate with interfaces yet,
@@ -286,6 +292,15 @@ gnome_file_entry_class_init (GnomeFileEntryClass *class)
 						   "or query any of its parameters."),
 						 GTK_TYPE_ENTRY,
 						 G_PARAM_READABLE));
+	g_object_class_install_property (gobject_class,
+					 PROP_USE_FILECHOOSER,
+					 g_param_spec_boolean (
+						 "use_filechooser",
+						 _("Use GtkFileChooser"),
+						 _("Whether to use the new GtkFileChooser widget "
+						   "or the GtkFileSelection widget to select files."),
+						 FALSE,
+						 G_PARAM_READWRITE));
 
 	class->browse_clicked = browse_clicked;
 	class->activate = NULL;
@@ -328,6 +343,9 @@ fentry_set_property (GObject *object, guint param_id,
 		gnome_file_entry_set_default_path (fentry, g_value_get_string (value));
 		break;
 
+	case PROP_USE_FILECHOOSER:
+		priv->use_filechooser = g_value_get_boolean (value);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
 		break;
@@ -382,6 +400,10 @@ fentry_get_property (GObject *object, guint param_id,
 		g_value_set_object (value, gnome_file_entry_gtk_entry (fentry));
 		break;
 
+	case PROP_USE_FILECHOOSER:
+		g_value_set_boolean (value, priv->use_filechooser);
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
 		break;
@@ -391,24 +413,37 @@ fentry_get_property (GObject *object, guint param_id,
 static void
 browse_dialog_ok (GtkWidget *widget, gpointer data)
 {
-	GtkFileSelection *fs;
+	GtkWidget *fw;
 	GnomeFileEntry *fentry;
 	GtkWidget *entry;
 	const gchar *locale_filename;
 	gchar *utf8_filename;
 
-	fs = GTK_FILE_SELECTION (data);
-	fentry = GNOME_FILE_ENTRY (g_object_get_data (G_OBJECT (fs), "gnome_file_entry"));
+	fw = GTK_WIDGET (data);
+	fentry = GNOME_FILE_ENTRY (g_object_get_data (G_OBJECT (fw), "gnome_file_entry"));
 	entry = gnome_file_entry_gtk_entry (fentry);
 
-	locale_filename = gtk_file_selection_get_filename (fs);
+	if (GTK_IS_FILE_CHOOSER (fentry->fsw))
+		locale_filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (fw));
+	else
+		locale_filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (fw));
+
 	utf8_filename = g_filename_to_utf8 (locale_filename, -1, NULL,
 					    NULL, NULL);
 	gtk_entry_set_text (GTK_ENTRY (entry), utf8_filename);
 	g_free (utf8_filename);
 	/* Is this evil? */
 	g_signal_emit_by_name (entry, "changed");
-	gtk_widget_destroy (GTK_WIDGET (fs));
+	gtk_widget_destroy (fw);
+}
+
+static void
+browse_dialog_response (GtkWidget *widget, gint response, gpointer data)
+{
+	if (response == GTK_RESPONSE_ACCEPT)
+		browse_dialog_ok (widget, data);
+	else
+		gtk_widget_destroy (GTK_WIDGET (data));
 }
 
 static void
@@ -507,66 +542,123 @@ build_filename (GnomeFileEntry *fentry)
 static void
 browse_clicked(GnomeFileEntry *fentry)
 {
-	GtkFileSelection *fs;
+	GtkWidget *fw;
 	char *p;
 	GtkWidget *toplevel;
 	GClosure *closure;
 
 	/*if it already exists make sure it's shown and raised*/
-	if(fentry->fsw) {
-		gtk_widget_show(fentry->fsw);
-		if(fentry->fsw->window)
-			gdk_window_raise(fentry->fsw->window);
-		fs = GTK_FILE_SELECTION(fentry->fsw);
-		gtk_widget_set_sensitive(fs->file_list,
+	if (fentry->fsw) {
+		if (GTK_IS_FILE_CHOOSER (fentry->fsw)) {
+			fw = fentry->fsw;
+		
+			gtk_widget_show (fw);
+
+			if (fw->window)
+				gdk_window_raise (fw->window);
+
+			gtk_file_chooser_set_folder_mode (GTK_FILE_CHOOSER (fw),
+							  fentry->_priv->directory_entry);
+
+			p = build_filename (fentry);
+			if (p) {
+				gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (fw), p);
+				g_free (p);
+			}
+
+			return;
+		} else {
+			fw = fentry->fsw;
+		
+			gtk_widget_show(fw);
+			if(fw->window)
+				gdk_window_raise(fw->window);
+
+			gtk_widget_set_sensitive(GTK_FILE_SELECTION (fw)->file_list,
+						 ! fentry->_priv->directory_entry);
+
+			p = build_filename (fentry);
+			if (p) {
+				gtk_file_selection_set_filename (GTK_FILE_SELECTION (fw), p);
+				g_free (p);
+			}
+			
+			return;
+		}
+	}
+
+	/* doesn't exist, create it */
+	if (fentry->_priv->use_filechooser) {
+		fentry->fsw = gtk_file_chooser_dialog_new (fentry->_priv->browse_dialog_title
+							   ? fentry->_priv->browse_dialog_title
+							   : _("Select file"),
+							   NULL,
+							   GTK_FILE_CHOOSER_ACTION_OPEN,
+							   GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+							   GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
+							   NULL);
+		fw = fentry->fsw;
+
+		gtk_dialog_set_default_response (GTK_DIALOG (fw), GTK_RESPONSE_ACCEPT);
+
+		gtk_file_chooser_set_folder_mode (GTK_FILE_CHOOSER (fw),
+						  fentry->_priv->directory_entry);
+
+		p = build_filename (fentry);
+		if (p) {
+			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (fw), p);
+			g_free (p);
+		}
+
+		closure = g_cclosure_new (G_CALLBACK (browse_dialog_response), fw, NULL);
+		g_object_watch_closure (G_OBJECT (fw), closure);
+
+		g_signal_connect_closure_by_id (fw,
+						g_signal_lookup ("response", G_OBJECT_TYPE (GTK_FILE_CHOOSER (fw))),
+						0, closure, FALSE);
+	} else {
+		fentry->fsw = gtk_file_selection_new (fentry->_priv->browse_dialog_title
+						      ? fentry->_priv->browse_dialog_title
+						      : _("Select file"));
+
+		fw = fentry->fsw;
+		
+		gtk_widget_set_sensitive(GTK_FILE_SELECTION (fw)->file_list,
 					 ! fentry->_priv->directory_entry);
 
 		p = build_filename (fentry);
-		if (p != NULL) {
-			gtk_file_selection_set_filename (fs, p);
+		if (p) {
+			gtk_file_selection_set_filename (GTK_FILE_SELECTION (fw), p);
 			g_free (p);
 		}
-		return;
+
+		closure = g_cclosure_new (G_CALLBACK (browse_dialog_ok), fw, NULL);
+		g_object_watch_closure (G_OBJECT (fw), closure);
+
+		g_signal_connect_closure_by_id (GTK_FILE_SELECTION (fw)->ok_button,
+						g_signal_lookup ("clicked", G_OBJECT_TYPE (GTK_FILE_SELECTION (fw)->ok_button)),
+						0, closure, FALSE);
+
+		g_signal_connect_swapped (GTK_FILE_SELECTION (fw)->cancel_button, "clicked",
+					  G_CALLBACK (gtk_widget_destroy),
+					  fw);
 	}
 
-
-	fentry->fsw = gtk_file_selection_new (fentry->_priv->browse_dialog_title
-					      ? fentry->_priv->browse_dialog_title
-					      : _("Select file"));
-
-	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (fentry));
-	if (GTK_WIDGET_TOPLEVEL (toplevel) && GTK_IS_WINDOW (toplevel))
-		gtk_window_set_transient_for (GTK_WINDOW (fentry->fsw), GTK_WINDOW(toplevel));
-
-	g_object_set_data (G_OBJECT (fentry->fsw), "gnome_file_entry", fentry);
-
-	fs = GTK_FILE_SELECTION (fentry->fsw);
-	gtk_widget_set_sensitive(fs->file_list,
-				 ! fentry->_priv->directory_entry);
-
-	p = build_filename (fentry);
-	if (p != NULL) {
-		gtk_file_selection_set_filename (fs, p);
-		g_free (p);
-	}
-
-	closure = g_cclosure_new (G_CALLBACK (browse_dialog_ok), fs, NULL);
-	g_object_watch_closure (G_OBJECT (fentry->fsw), closure);
-	g_signal_connect_closure_by_id (fs->ok_button,
-					g_signal_lookup ("clicked", G_OBJECT_TYPE (fs->ok_button)),
-					0, closure, FALSE);
-
-	g_signal_connect_swapped (fs->cancel_button, "clicked",
-				  G_CALLBACK (gtk_widget_destroy),
-				  fentry->fsw);
-	g_signal_connect (fentry->fsw, "destroy",
+	g_signal_connect (fw, "destroy",
 			  G_CALLBACK (browse_dialog_kill),
 			  fentry);
 
-	if (fentry->_priv->is_modal)
-		gtk_window_set_modal (GTK_WINDOW (fentry->fsw), TRUE);
+	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (fentry));
 
-	gtk_widget_show (fentry->fsw);
+	if (GTK_WIDGET_TOPLEVEL (toplevel) && GTK_IS_WINDOW (toplevel))
+		gtk_window_set_transient_for (GTK_WINDOW (fw), GTK_WINDOW(toplevel));
+
+	g_object_set_data (G_OBJECT (fw), "gnome_file_entry", fentry);
+
+	if (fentry->_priv->is_modal)
+		gtk_window_set_modal (GTK_WINDOW (fw), TRUE);
+
+	gtk_widget_show (fw);
 }
 
 static void
